/* create a dialog for processing standard input/output with wegina wexx
 * Copyright 2001,2003 Patrick TJ McPhee
 * Released under the Mozilla Public licence
 * The original code is wegina
 * The original author is Patrick McPhee
 * $Header: C:/ptjm/rexx/wegina/RCS/wegdlg.c 1.8 2003/03/31 03:16:35 ptjm Rel $
 */

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>

static struct sayDlg {
   HWND hwnd;
   HANDLE createSem, pullSem;
   int xpos, ypos;
   wchar_t * title;
   char * posName;
} sayWnd = { NULL, NULL, NULL, 0, 0, L"Wegina Wexx", "Position" },
traceWnd = { NULL, NULL, NULL, 0, 0, L"Wegina Twace", "TwacePosition" };

static int OKtoClose = 0;

#define IDSAY  100
#define IDPULL 101

#define SD_SAY  (WM_USER+1)
#define SD_LLUP (WM_USER+2)
#define SD_PULL (WM_USER+3)

/* control IDs that MS doesn't bother to define in their toolkit */
#define CC_BUTTON    0x0080 /* Button == L"Button" */
#define CC_EDIT      0x0081 /* Edit box == L"Edit" */
#define CC_STATIC    0x0082 /* Static text == L"Static" */
#define CC_LISTBOX   0x0083 /* List box == L"List" ? */
#define CC_SCROLLBAR 0x0084 /* Scroll bar == L"Scroll" ? */
#define CC_COMBOBOX  0x0085 /* Combo box == L"Combo" ? */


static BOOL CALLBACK sayDlg(HWND hwnd, UINT msg, WPARAM p1, LPARAM p2)
{
   switch (msg) {
      case WM_COMMAND:
         if (LOWORD(p1) == IDOK && OKtoClose) {
            PostMessage(hwnd, WM_CLOSE, 0, 0);
            return TRUE;
         }
         break;
      case WM_CLOSE:
            if (OKtoClose) {
               WINDOWPLACEMENT wpos;
               struct sayDlg * dlg = (struct sayDlg *)GetWindowLong(hwnd, GWL_USERDATA);

               /* remember where we were, because it's really irritating if
                * you don't */
               GetWindowPlacement(hwnd, &wpos);

               dlg->xpos = wpos.rcNormalPosition.left;
               dlg->ypos = wpos.rcNormalPosition.top;
               DestroyWindow(hwnd);
            }
            return TRUE;
            break;
      case SD_SAY: {
         static int len;

         SendDlgItemMessage(hwnd, IDSAY, EM_SETSEL, len, len);
         SendDlgItemMessage(hwnd, IDSAY, EM_REPLACESEL, FALSE, p2);
         len += p1;
         SendDlgItemMessage(hwnd, IDSAY, EM_SETSEL, len, len);
         SendDlgItemMessage(hwnd, IDSAY, EM_REPLACESEL, FALSE, (long)"\r\n");
         len += 2;
         SendDlgItemMessage(hwnd, IDSAY, EM_SETSEL, len-2, len-2);
         SendDlgItemMessage(hwnd, IDSAY, EM_SCROLLCARET, 0, 0);

         return TRUE;
      }
         break;
      case SD_LLUP:
         SetFocus(GetDlgItem(hwnd, IDOK));
         EnableWindow(GetDlgItem(hwnd, IDPULL), 0);
         GetDlgItemText(hwnd, IDPULL, (char *)p2, p1);
         SendDlgItemMessage(hwnd, IDPULL, WM_SETTEXT, 0, (long)"");
         return TRUE;
         break;
      case SD_PULL:
         EnableWindow(GetDlgItem(hwnd, IDPULL), 1);
         SetFocus(GetDlgItem(hwnd, IDPULL));
         return TRUE;
         break;
   }

   return FALSE;
}

void getPos(const char * name, int * xpos, int * ypos)
{
   char buf[1024];
   int bufl = sizeof(buf);
   HKEY hkey;
   int rc, type;
   
   rc = RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\ptjm\\wegina", 0, KEY_READ, &hkey);

   if (rc == ERROR_SUCCESS) {
      rc = RegQueryValueEx(hkey, name, NULL, &type, buf, &bufl);

      if (rc == ERROR_SUCCESS) {
         sscanf(buf, "%d %d", xpos, ypos);
      }
      RegCloseKey(hkey);
   }
}

void savePos(const char * name, int xpos, int ypos)
{
   char buf[1024];
   int bufl = sizeof(buf);
   HKEY hkey;
   int rc, type, result;
   
   rc = RegCreateKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\ptjm\\wegina", 0, "",
                       REG_OPTION_NON_VOLATILE, MAXIMUM_ALLOWED, NULL, &hkey, &result);

   if (rc == ERROR_SUCCESS) {
      bufl = sprintf(buf, "%d %d", xpos, ypos) + 1;
      rc = RegSetValueEx(hkey, name, 0, REG_SZ, buf, bufl);
      RegCloseKey(hkey);
   }
}




typedef struct ptjdimen {
   int x, y, cx, cy;
} * ptjdimen;


typedef struct ptjdialog {
   int alloc, used;
   unsigned char * next;
   DLGTEMPLATE tplt;
} * ptjdialog;

ptjdialog new_dialog(wchar_t * title, int style, ptjdimen dimen)
{
   ptjdialog pd;
   wchar_t * m;

   pd = malloc(sizeof(*pd)+1024);
   if (pd) {
      pd->alloc = 1024;
      pd->used = wcslen(title) * 2 + 2;
      m = (wchar_t *)(&pd->tplt+1);
      *m++ = 0;      /* no menu */
      *m++ = 0;      /* no dialog box class */
      memcpy(m, title, pd->used);

      pd->next = (char *)m + pd->used;
      pd->used += 4;
      
      /* align on dword boundary */
      while ((pd->used+2) & 0x3) {
         pd->next ++;
         pd->used ++;
      }

      pd->tplt.style = style;
      pd->tplt.dwExtendedStyle = 0;
      pd->tplt.cdit = 0;

      if (dimen) {
         pd->tplt.x = dimen->x;
         pd->tplt.y = dimen->y;
         pd->tplt.cx = dimen->cx;
         pd->tplt.cy = dimen->cy;
      }
      else {
         pd->tplt.x = pd->tplt.y = CW_USEDEFAULT;
         pd->tplt.cx = 300;
         pd->tplt.cy = 100;
      }
   }

   return pd;
}





ptjdialog add_control(ptjdialog dlg, wchar_t * cls, wchar_t * title, int id,
                      int style, ptjdimen dimen)
{
   DLGITEMTEMPLATE * ctrl;
   int touse;                          /* memory needed for this control */
   wchar_t *m;
   
   /* If cls starts with 0xffff, the second short is a numeric identifier
    * for the control. Otherwise, cls is a null-terminated Unicode string. */
   if (*cls == 0xffff) {
      touse = 2;
   }
   else {
      touse = (wcslen(cls) + 1) * 2;
   }

   /* title is a null-terminated Unicode string. The extra 2 bytes are for
    * the creation data, which I'm not supporting just now (so it ends up
    * as a zero-length Unicode string) (which has length 2 for the null
    * terminator) */
   touse += (wcslen(title) + 1) * 2 + 2 + sizeof(*ctrl);

   /* each dialog item needs to be word-aligned */
   while (touse & 0x3) touse++;

   if ((dlg->used + touse) > dlg->alloc) {
      dlg = realloc(dlg, dlg->alloc + touse + 1024);
      dlg->alloc += touse + 1024;
      dlg->next = (char *)(dlg+1) + dlg->used;
   }

   ctrl = (DLGITEMTEMPLATE *)dlg->next;

   dlg->used += touse;
   dlg->next += touse;

   ctrl->style = style;
   ctrl->dwExtendedStyle = 0;

   if (dimen) {
      ctrl->x = dimen->x;
      ctrl->y = dimen->y;
      ctrl->cx = dimen->cx;
      ctrl->cy = dimen->cy;
   }
   else {
      ctrl->x = 0;
      ctrl->y = 0;
      ctrl->cx = 40;
      ctrl->cy = 10;
   }

   ctrl->id = id;

   m = (wchar_t *)(ctrl + 1);

   if (*cls == 0xffff) {
      memcpy(m, cls, 4);
      m += 2;
   }
   else {
      touse = wcslen(cls) + 1;
      memcpy(m, cls, touse*2);
      m += touse;
   }

   touse = wcslen(title) + 1;
   memcpy(m, title, touse*2);
   m += touse;
   *m++ = 0;
   dlg->tplt.cdit++;

   return dlg;
}


static void createSayDlg(void * arg)
{
   ptjdialog tplt;
   MSG msg;
   struct ptjdimen d;
   HWND okButton, pullWindow, sayWindow;
   HICON defIcon;
   extern HINSTANCE myinstance;
   struct sayDlg * dlg = (struct sayDlg *)arg;
   int rc;

   d.x = dlg->xpos;
   d.y = dlg->ypos;
   d.cx = 300;
   d.cy = 100;

   tplt = new_dialog(dlg->title, WS_CAPTION|WS_VISIBLE|DS_3DLOOK|WS_SYSMENU|
                     WS_DLGFRAME|WS_MINIMIZEBOX, &d);

   d.x = 250;
   d.y = 80;
   d.cx = 40;
   d.cy = 10;

   tplt = add_control(tplt, L"Button", L"Close", IDOK, WS_VISIBLE|BS_DEFPUSHBUTTON|
                      WS_TABSTOP, &d);

   d.x = 10;
   d.y = 10;
   d.cx = 280;
   d.cy = 60;

   tplt = add_control(tplt, L"Edit", L"", IDSAY, WS_VISIBLE|WS_VSCROLL|WS_HSCROLL|ES_READONLY|
                      ES_MULTILINE|ES_AUTOVSCROLL|ES_AUTOHSCROLL|WS_BORDER|WS_TABSTOP, &d);

   d.x = 10;
   d.y = 80;
   d.cx = 220;
   d.cy = 10;

   tplt = add_control(tplt, L"Edit", L"", IDPULL, WS_VISIBLE|ES_AUTOHSCROLL|
                      WS_BORDER|WS_DISABLED|WS_TABSTOP, &d);

   defIcon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(1));

   dlg->hwnd =  CreateDialogIndirect(GetModuleHandle(NULL), &tplt->tplt, HWND_DESKTOP, sayDlg);
   if (!dlg->hwnd)
      rc = GetLastError();

   free(tplt);

   SetEvent(dlg->createSem);

   if (dlg->hwnd == NULL) {
      static const char msgprefix[] = "Failed to create dialog box: ";
      char buf[1024];

      memcpy(buf, msgprefix, sizeof(msgprefix)-1);
      FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(),
                    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), buf+sizeof(msgprefix)-1,
                    sizeof(buf) - sizeof(msgprefix), NULL);

      MessageBox(HWND_DESKTOP, buf, "Dwat!", MB_OK);
   }

   getPos(dlg->posName, &dlg->xpos, &dlg->ypos);
   SetWindowPos(dlg->hwnd, NULL, dlg->xpos, dlg->ypos, 0, 0,
                SWP_NOSENDCHANGING|SWP_NOSIZE|SWP_NOZORDER);

   PostMessage(dlg->hwnd, WM_SETICON, ICON_BIG, (long)defIcon);
   SetWindowLong(dlg->hwnd, GWL_USERDATA, (long)dlg);

   okButton = GetDlgItem(dlg->hwnd, IDOK);
   pullWindow = GetDlgItem(dlg->hwnd, IDPULL);
   sayWindow = GetDlgItem(dlg->hwnd, IDSAY);

   EnableWindow(okButton, 0);


   while (GetMessage(&msg, dlg->hwnd, 0, SD_PULL) > 0) {
      TranslateMessage(&msg);

      /* handle tabs */
      if (msg.message == WM_KEYDOWN) {
         if (msg.wParam == 0x9) {
            if (msg.hwnd == okButton) {
               if (!SetFocus(pullWindow))
                  SetFocus(sayWindow);
            }
            else
               SetFocus(okButton);
         }
         else if (msg.wParam == 0xd && msg.hwnd == okButton) {
            PostMessage(msg.hwnd, WM_LBUTTONDOWN, 0, 0);
            PostMessage(msg.hwnd, WM_LBUTTONUP, 0, 0);
         }
         else if (msg.wParam == 0xd && msg.hwnd == pullWindow) {
            SetEvent(dlg->pullSem);
         }
      }

      DispatchMessage(&msg);
   }

   savePos(dlg->posName, dlg->xpos, dlg->ypos);

   SetEvent(dlg->createSem);
}

static void launchDlg(struct sayDlg * dlg)
{
      /* event will be signalled when window is created  */
      dlg->createSem = CreateEvent(NULL, TRUE, FALSE, NULL);
      dlg->pullSem = CreateEvent(NULL, TRUE, FALSE, NULL);
      _beginthread(createSayDlg, 102400, dlg);

      WaitForSingleObject(dlg->createSem, INFINITE);

      /* it will be signalled again when the window is destroyed */
      ResetEvent(dlg->createSem);
}


void say(const char * s, const unsigned long l)
{
   if (sayWnd.hwnd == NULL) {
      launchDlg(&sayWnd);
   }
   
   SendMessage(sayWnd.hwnd, SD_SAY, l, (long)s);
}

void saytrace(const char * s, const unsigned long l)
{
   if (traceWnd.hwnd == NULL) {
      launchDlg(&traceWnd);
   }
   
   SendMessage(traceWnd.hwnd, SD_SAY, l, (long)s);
}

static void Pull(struct sayDlg * dlg, char * s, unsigned long * l)
{
   if (dlg->hwnd == NULL) {
      launchDlg(dlg);
   }
   
   PostMessage(dlg->hwnd, SD_PULL, 0, 0);
   WaitForSingleObject(dlg->pullSem, INFINITE);
   ResetEvent(dlg->pullSem);

   SendMessage(dlg->hwnd, SD_LLUP, *l, (long)s);
   *l = strlen(s);
}



void pull(char * s, unsigned long * l)
{
   Pull(&sayWnd, s, l);
}

void pulltrace(char * s, unsigned long * l)
{
   Pull(&traceWnd, s, l);
}

static void wait_for_it(struct sayDlg * dlg)
{
   if (dlg->hwnd) {
      WaitForSingleObject(dlg->createSem, INFINITE);
      dlg->hwnd = NULL;
      CloseHandle(dlg->createSem);
      CloseHandle(dlg->pullSem);
   }
}


void wait_for_say()
{
   OKtoClose = 1;
   if (sayWnd.hwnd) {
      EnableWindow(GetDlgItem(sayWnd.hwnd, IDOK), 1);
   }
   if (traceWnd.hwnd) {
      EnableWindow(GetDlgItem(traceWnd.hwnd, IDOK), 1);
   }


   wait_for_it(&sayWnd);
   wait_for_it(&traceWnd);
}
